"""
Index search.

"""
from elasticsearch_dsl.mapping import Mapping

from microcosm_elasticsearch.errors import translate_elasticsearch_errors


class SearchIndex:
    """
    Encapsulates search against an index.

    An index may have many (polymorphic) document types registered with it. The search index
    handles this complexity in two ways:

     -  It creates instances of a specific model class as the results of any searches.

        If searching a polymorphic index, the model class should be a compatible base class.

     -  It can restrict the search to specific document types.

    """
    __mapping_type_name__ = "doc"

    @property
    def mapping_type_name(self):
        return self.__class__.__mapping_type_name__

    @property
    def doc_type_field(self):
        """
        Defines document field used to determine the document's polymorphic type in a single-mapping-type index

        """
        return "doctype"

    def __init__(self, graph, index, doc_type=None):
        """
        :param graph: the object graph
        :param index: the name of an index to use
        :param doc_type: the doc type to search for; uses the index's doc types if omitted

        """
        self.elasticsearch_client = graph.elasticsearch_client
        self.index = index
        # Mapping from ES custom type field to corresponding model class
        self.doc_types = dict()
        self.mapping = Mapping(self.mapping_type_name)
        if doc_type is not None:
            self.register_doc_type(doc_type)

    def register_doc_type(self, model_class):
        model_doctype = model_class.get_model_doctype()
        self.doc_types[model_doctype] = model_class
        self.update_mapping(model_class)

    def update_mapping(self, model_class):
        self.mapping.update(model_class._doc_type.mapping)
        self.index.mapping(self.mapping)

    @property
    def index_name(self):
        return self.index._name

    @translate_elasticsearch_errors
    def count(self, **kwargs):
        """
        Count the number of models matching some criterion.

        """
        query = self._search(**kwargs)
        return query.count()

    @translate_elasticsearch_errors
    def search(self, **kwargs):
        """
        Return the list of models matching some criterion.

        :param offset: pagination offset, if any
        :param limit: pagination limit, if any

        """
        query = self._search(**kwargs)
        results = query.execute()
        return self._to_list(results)

    def search_with_count(self, **kwargs):
        """
        Return the list of models matching some criterion.

        :param offset: pagination offset, if any
        :param limit: pagination limit, if any

        """
        query = self._search(**kwargs)
        results = query.execute()
        return self._to_list(results), query.count()

    def _search(self, explain=False, **kwargs):
        query = self._query()
        query = self._order_by(query, **kwargs)
        query = self._filter(query, **kwargs)
        if explain:
            query = query.extra(explain=True)
        return query

    def _to_instance(self, hit):
        """
        Resolve this hit into a model instance.

        """
        hit_doc_type = getattr(hit, self.doc_type_field, None)
        hit_model_class = self.doc_types.get(hit_doc_type)
        if hit_doc_type is not None and hit_model_class is not None:
            return hit_model_class(meta=hit.meta.to_dict(), **hit._d_)
        # Will return be a generic `Hit`
        return hit

    def _to_list(self, results):
        """
        Resolve this hit into a list of instances.

        """
        return [
            self._to_instance(hit)
            for hit in results.hits
        ]

    def _query(self):
        """
        Create a search query.

        Starts with the index's search function; customizes with the provided doc types, if any

        """
        return self.index.search()

    def _order_by(self, query, **kwargs):
        """
        Add an order by clause to a (search) query.

        By default, uses reverse chronogical creation order.

        """
        return query.sort("-created_at")

    def _filter(self, query, offset=None, limit=None, doc_type=None, doc_types=(), **kwargs):
        """
        Filter a query with user-supplied arguments.

        :param doc_types: a list of legal doc types
        :param offset: pagination offset, if any
        :param limit: pagination limit, if any

        """
        if doc_type or doc_types:
            query = query.filter(
                "terms",
                **{
                    self.doc_type_field: [
                        doc_type_
                        for doc_type_ in self._iter_doc_types(doc_type, doc_types)
                    ],
                },
            )

        if offset is not None:
            query = query.extra(from_=offset)

        if limit is not None:
            query = query.extra(size=limit)

        return query

    def _iter_doc_types(self, doc_type=None, doc_types=()):
        if doc_type:
            yield doc_type
        yield from doc_types



"""
Synthetic object containing information about an index.

"""


class IndexStatus:
    def __init__(
        self,
        name,
        data,
        stats,
    ):
        self.name = name
        self.data = data
        self.stats = stats



"""
Base classes for models

"""
from elasticsearch_dsl import Date, Document, Keyword


class Model(Document):
    """
    Base for all models. Any model instance has a primary key and created/updated timestamps,
    as well as a field indicating the doctype.

    See README#polymorphic-models for more details

    """
    # Every persistent entity should have a primary key id and created/updated timestamps.
    id = Keyword(required=True)
    created_at = Date(required=True)
    updated_at = Date(required=True)

    # In ES6+ indices have a single mapping type
    # To allow for polymorphic documents, we define a field holding the document type
    doctype = Keyword(required=True)

    def __init__(self, meta=None, **kwargs):
        cls = self.__class__
        kwargs["doctype"] = cls.get_model_doctype()
        return super().__init__(meta=meta, **kwargs)

    @classmethod
    def get_model_doctype(cls):
        return getattr(cls, "__doctype_name__", None) or cls.__name__.lower()

    def _members(self):
        return {
            key: value
            for key, value in self.to_dict().items()
            # NB: exclude non-persistent fields
            if key not in ("index", "doc_type")
        }

    def __eq__(self, other):
        return type(other) is type(self) and self._members() == other._members()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self) if self.id is None else hash(self.id)

    @property
    def _id(self):
        return self.meta.id

    @_id.setter
    def _id(self, value):
        self.meta.id = value

    @property
    def _index(self):
        return self.meta.index

    @_index.setter
    def _index(self, value):
        self.meta.index = value

    @property
    def _score(self):
        return self.meta.score

    @_score.setter
    def _score(self, value):
        self.meta.score = value

    @property
    def _version(self):
        return self.meta.version

    @_version.setter
    def _version(self, value):
        self.meta.version = value



"""
Index Status convention.

"""
from marshmallow import Schema
from microcosm_flask.conventions.base import EndpointDefinition
from microcosm_flask.conventions.crud import configure_crud
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation

from microcosm_elasticsearch.index_status.resources import IndexStatusSchema
from microcosm_elasticsearch.index_status.store import IndexStatusStore


def configure_status_convention(graph):
    store = IndexStatusStore(graph)

    ns = Namespace(
        subject="index_status",
    )

    def search(**kwargs):
        status = store.get_status()
        return status, len(status)

    mappings = {
        Operation.Search: EndpointDefinition(
            func=search,
            request_schema=Schema(),
            response_schema=IndexStatusSchema(),
        ),
    }

    configure_crud(graph, ns, mappings)
    return ns



"""
Some useful decorators

"""
from functools import wraps

from elasticsearch.exceptions import AuthorizationException
from microcosm_flask.errors import as_retryable


def retryable(func):
    """
    Mark authorization error as retryable

    Useful e.g. when using an AWS-hosted ES
    AWS credentials can expire; encourage retries for synchronous requests.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AuthorizationException as error:
            raise as_retryable(error)
    return wrapper



"""
Factory that configures Elasticsearch client.

"""
from functools import partial
from os import environ
from urllib.parse import parse_qs, urlencode, urlparse

from boto3 import Session
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from elasticsearch import Elasticsearch, RequestsHttpConnection
from microcosm.api import defaults
from microcosm.config.types import boolean
from microcosm.config.validation import typed


def make_url_safe(raw_url):
    """
    Make sure all query parameters are URL encoded

    Some servers can deal with some parameters not being
    strictly urlencoded, but some canonization inside
    the signing logic means that we have to pre-urlencode
    all query parameters.
    """
    url = urlparse(raw_url)
    path = url.path or "/"
    if url.query:
        querystring = "?" + urlencode(parse_qs(url.query,
                                               keep_blank_values=True),
                                      doseq=True)
    else:
        querystring = ""
    return (url.scheme +
            "://" +
            url.netloc +
            path +
            querystring)


def awsv4sign(r, *, session, region):
    request = AWSRequest(method=r.method.upper(),
                         url=make_url_safe(r.url),
                         data=r.body)
    credentials = session.get_credentials()
    SigV4Auth(credentials, 'es', region).add_auth(request)
    r.headers.update(dict(request.headers.items()))
    return r


@defaults(
    aws_region=environ.get("AWS_DEFAULT_REGION", environ.get("AWS_REGION", "us-east-1")),
    host="localhost",
    # NB: these are the defaults shipped with the ES docker distribution.
    # We want testing to "just work"; no sane production application should use these.
    username="elastic",
    password="changeme",
    use_aws4auth=typed(boolean, default_value=False),
    timeout_seconds=typed(int, 10),
    retry_on_timeout=typed(boolean, default_value=False),
)
def configure_elasticsearch_client(graph):
    """
    Configure Elasticsearch client using a constructed dictionary config.

    :returns: an Elasticsearch client instance of the configured name

    """
    if graph.config.elasticsearch_client.use_aws4auth:
        region = graph.config.elasticsearch_client.aws_region
        awsauth = partial(
            awsv4sign,
            session=Session(),
            region=region,
        )
        config = dict(
            hosts=[{
                "host": graph.config.elasticsearch_client.host,
                "port": 443,
            }],
            connection_class=RequestsHttpConnection,
            http_auth=awsauth,
            use_ssl=True,
            verify_certs=True,
            timeout=graph.config.elasticsearch_client.timeout_seconds,
            retry_on_timeout=graph.config.elasticsearch_client.retry_on_timeout,
        )
    else:
        config = dict(
            hosts=[
                graph.config.elasticsearch_client.host,
            ],
            http_auth=(
                graph.config.elasticsearch_client.username,
                graph.config.elasticsearch_client.password,
            ),
        )
    return Elasticsearch(**config)



"""
Abstraction layer for persistence operations.

Allows upstream (e.g. HTTP) operations to use a consistent persistence interface
and avoids leaking persistence abstractions throughout a code base.

Intended to be duck-type compatible with `microcosm_postgres.store.Store`.

"""
from contextlib import contextmanager
from time import time
from uuid import uuid4

from elasticsearch.helpers import bulk

from microcosm_elasticsearch.errors import translate_elasticsearch_errors


class Store:
    """
    Elasticsearch persistence interface.

    """
    def __init__(self, graph, index, model_class, search_index):
        """
        :param graph: the object graph
        :param index: the name of an index to use
        :param model_class: a `elasticsearch_dsl.Document` subclass to persist.

        """
        self.elasticsearch_client = graph.elasticsearch_client
        self.index = index
        self.model_class = model_class

        search_index.register_doc_type(model_class)
        self.search_index = search_index

        # NB: do NOT provide a model backref here because "smart" shortcuts on the
        # model will conflict with existing methods on the DocType base class

    @property
    def doc_type(self):
        return self.model_class._doc_type.name

    @property
    def index_name(self):
        return self.index._name

    @contextmanager
    def flushing(self):
        """
        Flush the current session if there's no error.

        Flushing an index is not an expected behavior for Elasticsearch writes, but
        can be very useful for test cases.

        """
        yield
        self.index.flush()

    def new_object_id(self):
        """
        Injectable id generation to facilitate mocking.

        Uses string-valued UUID because ES uuid support is not terrific.

        """
        return str(uuid4())

    def new_timestamp(self):
        """
        Injectable timestamp generation to facilitate mocking.

        """
        # NB: Elasticsearch supports epoch_millis by default
        return int(time() * 1000)

    def count(self, **kwargs):
        # delegate
        return self.search_index.count(**kwargs)

    def search(self, **kwargs):
        # delegate
        return self.search_index.search(**kwargs)

    def search_with_count(self, **kwargs):
        # delegate
        return self.search_index.search_with_count(**kwargs)

    @translate_elasticsearch_errors
    def create(self, instance):
        """
        Persist an entity into Elasticsearch.

        """
        now_millis = self.new_timestamp()

        if instance.id is None:
            instance.id = self.new_object_id()

        instance._id = instance.id
        instance.created_at = now_millis
        instance.updated_at = now_millis

        # NB: the DSL save function will overwrite existing records; use the raw client
        self.elasticsearch_client.create(
            id=instance.id,
            index=self.index_name,
            doc_type=self.doc_type,
            body=instance.to_dict(),
        )
        return instance

    @translate_elasticsearch_errors
    def retrieve(self, identifier):
        """
        Retrieve a model by primary key and zero or more other criteria.

        :raises `ElasticsearchNotFoundError` if there is no existing model

        """
        return self.model_class.get(
            id=identifier,
            index=self.index_name,
            using=self.elasticsearch_client,
        )

    @translate_elasticsearch_errors
    def update(self, identifier, new_instance):
        """
        Update an existing model with a new one.

        :raises `ElasticsearchNotFoundError` if there is no existing model

        """
        new_instance.id = identifier
        new_instance._id = identifier
        new_instance.updated_at = self.new_timestamp()

        new_instance.update(
            using=self.elasticsearch_client,
            index=self.index_name,
            **new_instance.to_dict()
        )

        return self.retrieve(identifier)

    @translate_elasticsearch_errors
    def replace(self, identifier, new_instance):
        """
        Create or update an entity.

        """
        now_millis = self.new_timestamp()

        if new_instance.id is None:
            if identifier is None:
                identifier = self.new_object_id()
            new_instance.id = identifier

        if new_instance.created_at is None:
            new_instance.created_at = now_millis

        new_instance._id = new_instance.id
        new_instance.updated_at = now_millis

        new_instance.save(
            id=new_instance.id,
            using=self.elasticsearch_client,
            index=self.index_name,
            validate=True,
        )
        return new_instance

    @translate_elasticsearch_errors
    def delete(self, identifier):
        """
        Delete a model by primary key.

        :raises `ElasticsearchNotFoundError` if there is no existing model

        """
        self.model_class().delete(
            id=identifier,
            index=self.index_name,
            using=self.elasticsearch_client,
        )
        return True

    def _batch_bulk(self, actions, batch_size):
        """
        Breaks list of actions into batches

        """
        num_actions = len(actions)
        for offset in range(0, num_actions, batch_size):
            yield actions[offset:min(offset + batch_size, num_actions)]

    @translate_elasticsearch_errors
    def bulk(self, actions, batch_size):
        """
        Bulk index entities

        actions: list of tuples of (action, instance) to be included in the bulk
        batch_size: number of records for each bulk call

        All errors and exceptions are suppressed and are returned in the response report

        """
        def to_dict(instance, op_type):
            if instance.id is None:
                instance.id = self.new_object_id()

            instance._id = instance.id
            instance._index = self.index_name

            record = instance.to_dict(include_meta=True)
            if op_type == "delete":
                del record["_source"]

            record["_op_type"] = op_type
            return record

        actions = [
            to_dict(instance, op_type)
            for op_type, instance in actions
        ]
        return [
            bulk(
                client=self.elasticsearch_client,
                actions=actions_batch,
                index=self.index._name,
                raise_on_exception=False,
                raise_on_error=False,
            ) for actions_batch in self._batch_bulk(
                actions=actions,
                batch_size=batch_size,
            )
        ]



"""
Assertion support.

Elasticsearch inserts are not immediately indexed for search and count queries.

While it's possible to explicitly flush an index, the Elasticsearch persistence
model typically assumes that inserts are *not* indexed synchronously. For simpler
test code in these cases, it's helpful to have assertions that *eventually* resolve,
after some number of retries.

"""
from time import sleep

from hamcrest import assert_that, calling, raises


DEFAULT_TRIES = 3
DEFAULT_SLEEP_SECONDS = 0.1


def assert_that_eventually(func, matches, tries=DEFAULT_TRIES, sleep_seconds=DEFAULT_SLEEP_SECONDS):
    last_error = None
    for _ in range(tries):
        try:
            assert_that(func(), matches)
            break
        except Exception as error:
            last_error = error
        sleep(sleep_seconds)
    else:
        raise last_error


def assert_that_not_eventually(func, matches, tries=DEFAULT_TRIES, sleep_seconds=DEFAULT_SLEEP_SECONDS):
    assert_that(
        calling(assert_that_eventually).with_args(func, matches, tries, sleep_seconds),
        raises(AssertionError),
    )



from enum import Enum

from elasticsearch_dsl import Keyword
from elasticsearch_dsl.exceptions import ValidationException


class EnumField(Keyword):
    # Force serialization
    _coerce = True
    name = "keyword"

    def __init__(self, enum_class, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Leading underscore prevents attribute from being added to document in ElasticSearch
        if not issubclass(enum_class, Enum):
            raise TypeError("The class passed into an EnumField should a subclass of Enum")
        self._enum_class = enum_class

    def _serialize(self, data):
        return str(data)

    def _deserialize(self, data):
        return self._enum_class[data]

    def clean(self, data):
        """
        Field validation
        Handle both cases where the data comes in as a string or enum

        """
        if isinstance(data, str):
            return super().clean(data)
        elif isinstance(data, Enum):
            if data not in self._enum_class:
                raise ValidationException(f"Value should be a member of {self.enum_class}")
            return data
        else:
            raise ValidationException(f"Value of type {data.__class__} is not acceptable for an EnumField")



"""
Manage a set of indexes and/or aliases.

"""
from elasticsearch.exceptions import RequestError
from elasticsearch_dsl import Index
from inflection import underscore


class IndexRegistry:
    """
    A registry of application indexes.

    """
    def __init__(self, graph):
        self.graph = graph
        self.indexes = {}

    def register(self, name=None, version=None, settings=None):
        """
        Register an index locally.

        Note that `createall` is needed to save the index to Elasticsearch.

        The index will be named per convention such that:
         -  The graph's name is used by default
         -  The "test" suffix is added for unit testing (to avoid clobbering real data)

        If version is provided, it will be used to create generate an alias (to the unversioned name).

        """
        if version is None:
            index_name = IndexRegistry.name_for(self.graph, name=name)
            alias_name = None
        else:
            # create index with full version, alias to shortened version
            index_name = IndexRegistry.name_for(self.graph, name=name, version=version)
            alias_name = IndexRegistry.name_for(self.graph, name=name)

        if index_name in self.indexes:
            raise Exception("Index already registered for name: {}".format(index_name))

        index = Index(
            name=index_name,
            using=self.graph.elasticsearch_client,
        )

        if settings:
            index.settings(**settings)

        if alias_name is not None:
            index.aliases(**{alias_name: {}})

        self.indexes[index_name] = index
        return index

    def createall(self, force=False, only=(), skip=()):
        """
        Create all indexes in Elasticsearch.

        """
        only = set(only or [])
        skip = set(skip or [])

        for index in self.indexes.values():
            aliases = set(index._aliases)
            if only and (index._name not in only and not (aliases & only)):
                continue
            if skip and (index._name in skip or (aliases & skip)):
                continue
            if force and index.exists():
                index.delete()
                index.flush(ignore_unavailable=True)
            try:
                index.create()
                index.refresh()
            except RequestError as error:
                if "index_already_exists_exception" in str(error):
                    continue
                raise

    @staticmethod
    def name_for(graph, name=None, version=None):
        """
        Generate a convention-aware index name.

        In particular: uses a different index name for unit testing.

        """
        name_parts = [
            name or graph.metadata.name,
        ]

        if version is not None:
            name_parts.append(str(version))

        if graph.metadata.testing:
            name_parts.append("test")

        return "_".join([underscore(name_part) for name_part in name_parts])



"""
CLI entry point hook.

"""
from argparse import ArgumentParser
from json import loads


def createall_main(graph):
    """
    Initialize indexes and mappings.

    """
    parser = ArgumentParser()
    parser.add_argument("--only", action="append")
    parser.add_argument("--skip", action="append")
    parser.add_argument("-D", "--drop", action="store_true")
    args = parser.parse_args()

    graph.elasticsearch_index_registry.createall(
        force=args.drop,
        only=args.only,
        skip=args.skip,
    )


def query_main(graph, default_index):
    """
    Run a query.

    """
    parser = ArgumentParser()
    parser.add_argument("-i", "--index", default=default_index)
    parser.add_argument("-q", "--query", default='{"match_all": {}}')
    parser.add_argument("-f", "--flat", action="store_true")
    args = parser.parse_args()

    try:
        query = loads(args.query)
    except Exception:
        parser.error("query must be valid json")

    response = graph.elasticsearch_client.search(
        index=args.index,
        body=dict(query=query),
    )

    if args.flat:
        return response["hits"]["hits"]
    else:
        return response



"""
Index Status Store

"""
from microcosm_elasticsearch.index_status.models import IndexStatus


class IndexStatusStore:

    def __init__(self, graph):
        self.index_registry = graph.elasticsearch_index_registry

    def process_status_data(self, status, stats):
        """
        Iterate through returned data and return related stats

        """
        indices = []
        for name, data in status.items():
            indices.append(
                IndexStatus(
                    name=name,
                    data=data,
                    stats=stats['indices'][name]
                )
            )
        return indices

    def get_status(self):
        indices = []
        for key, index in self.index_registry.indexes.items():
            indices.extend(
                self.process_status_data(
                    status=index.get(),
                    stats=index.stats(),
                )
            )
        return indices



#!/usr/bin/env python
from setuptools import find_packages, setup


project = "microcosm-elasticsearch"
version = "6.2.0"

setup(
    name=project,
    version=version,
    description="Elasticsearch client configuration",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/microcosm-elasticsearch",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    keywords="microcosm",
    install_requires=[
        "boto3>=1.7.33",
        "elasticsearch>=6.3.1,<7.0.0",
        "elasticsearch-dsl>=6.2.1,<7.0.0",
        "microcosm>=2.4.0",
        "microcosm-flask>=1.13.1",
        "microcosm-metrics>=2.1.0",
        "requests[security]>=2.18.4",
        "urllib3<1.27",
    ],
    setup_requires=[
        "nose>=1.3.6",
    ],
    dependency_links=[
    ],
    entry_points={
        "microcosm.factories": [
            "elasticsearch_client = microcosm_elasticsearch.factories:configure_elasticsearch_client",
            "elasticsearch_index_registry = microcosm_elasticsearch.registry:IndexRegistry",
            "index_status_convention = microcosm_elasticsearch.index_status.convention:configure_status_convention",
        ],
    },
    tests_require=[
        "coverage>=4.4.1",
        "PyHamcrest>=1.9.0",
    ],
)



"""
Elasticsearch errors.

Compatible with `microcosm-flask` HTTP error handling conventions.

"""
from functools import wraps

from elasticsearch.exceptions import ConflictError, NotFoundError, RequestError


def translate_elasticsearch_errors(func):
    """
    Translate Elasticsearch errors into HTTP compatible ones.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ConflictError:
            raise ElasticsearchConflictError
        except NotFoundError:
            raise ElasticsearchNotFoundError
        except RequestError as error:
            raise ElasticsearchError(error)
        except KeyError as error:
            # NB: usually caused by a missing index argument
            raise ElasticsearchError(error)
    return wrapper


class ElasticsearchError(Exception):
    """
    Something unexpected happened.

    """
    @property
    def status_code(self):
        # internal server error
        return 500


class ElasticsearchConflictError(ElasticsearchError):
    """
    An attempt to create or update an entity violated a constraint.

    Often expected behavior.

    """
    @property
    def status_code(self):
        # conflict
        return 409

    @property
    def include_stack_trace(self):
        return False


class ElasticsearchNotFoundError(ElasticsearchError):
    """
    A supplied identifier does not refer to a known entity.

    """
    @property
    def status_code(self):
        # not found
        return 404

    @property
    def include_stack_trace(self):
        return False
